"""
This file is automitically generated during the setup.
Please do not edit config.py, but edit config.py.in in the root directory.
"""

MPI_INCLUDE_DIRS = "/home/Simone/anaconda3/include"
MPI_LINK_FLAGS = "-Wl,-rpath,/home/Simone/anaconda3/lib"
MPI_LIBRARIES = "/home/Simone/anaconda3/lib/libmpi.so"
